#!/usr/bin/perl -w

use strict 'vars';
use lib "../build_tools";
use Getopt::Std;
use pl_encoding;


my %options;

getopts('do:',\%options);

my $debug=$options{d} || 0;

if ($options{o}){
  open(STDOUT,">$options{o}") || die "$!";
}

sub cache_call(&\%@){
  my $subref=shift;
  my $cacheref=shift;
  my $key=join(",",@_);
  my $cacheval=${$cacheref}{$key};
  return  $cacheval if (defined($cacheval));

  my $rv= &$subref(@_);
  ${$cacheref}{$key}=$rv;
  return $rv;

}




our %mulorder=('c' => 1,
	   '0' => 2,
	   'a' => 3,
	   'u' => 4,
	   'n' => 5,
	   'p' => 6,
	   'g' => 7,
	   'l' => 8
	 );

our %plorder=('c' => 0,
	     '0' => 1,
	     'n' => 2,
	     'p' => 4,
	     'l' => 5,
	     'g' => 6,
	     'u' => 7,
	     'a' => 8
	 );

sub mulcmp {  $mulorder{$a} <=> $mulorder{$b} };

sub pl_cmp {  $plorder{$a} <=> $plorder{$b} };

our @signs=sort mulcmp keys (%sign_name);

my $bases=6;
my $encoding_width=3;
my $bitwidth=32;



my %mult1=('c'=>'c',
	'a'=>'a',
	'0'=>'0');

my %mult2=('u0' => '0',
	'up' => 'u',
	'un' => 'u',
	'uu' => 'u',
	'ul' => 'a',
	'ug' => 'a',
	'gg' => 'g',
	'll' => 'g',
	'gl' => 'l');

sub complement($){
  my $sign=shift;

  my @bits=split("",$sign);

  return join("",map( $_ ? 0 : 1 , @bits));
}

my %mult_cache=();
sub real_mult($$);
sub mult($$){ return cache_call(\&real_mult, %mult_cache, @_ )};

sub real_mult($$){
  my ($p,$q)=sort  mulcmp @_;
  my $pq=$p.$q;

  return $mult1{$p} if (defined($mult1{$p}));
  return $q if ($p eq 'p');
  return $neg{$q} if ($p eq 'n');
  return $mult2{$pq} if (defined($mult2{$pq}));
  
  die "unimplemented mult $pq";

}

sub fcol ($$$){
  my $forceto=shift;
  my $replace=shift;
  my $ts=shift;


  my @terms=split('',$ts);
  my @saveterms=@terms;

  
  my @indices=();
  my @mergein=split('',$forceto);
  my @merged=();

  my $i=0;
  for my $c (split('',$replace)){
    $i=index($ts,$c);
    die "cannot find $c in \'$ts\'" unless ($i>=0);
    substr($ts,$i,1)='@';

    my $m=shift(@mergein);
    my $mc=merge_sign($m,$c);
    return 'f' if ($mc eq 'c');
    next if ($mc eq $c);
    push(@indices,$i);
    push(@merged,$mc);
  }

  if (scalar(@indices)==0) {
    return 't';
  } else {
    return join('',@indices,':',@merged);
  }
}


sub pluecker_check_actual($$$);
my %pc_cache=();
sub pluecker_check(@){
  return cache_call(\&pluecker_check_actual,%pc_cache,@_);
}
sub pluecker_check_actual($$$){
  my $ts=join('',@_);
  my @terms=@_;


  my @sterms = sort pl_cmp @terms;
  local ($_)=join('',@sterms);

  #propagate contradictions
  /c/ && return 'f';

  # this set of signs is ok.	
  (/000/ || /np/) && return 't';

  # other fully determined sign combinations are contradictions		     
  ( /^[np0]+$/) && return 'f';

  # now consider cases where two signs are fully determined
  # note that one of the signs must be 0, or both equal
  (/0([0np])(.)/)	&& return fcol $neg{$1},$2,$ts;
  /nn(.)/ && return fcol 'p', $1, $ts;
  /pp(.)/ && return fcol 'n', $1, $ts;

  # everything <= 0 means everything = zero
  /([0l]{3}|[0g]{3})/ && return fcol '000',$1,$ts;

  /0gu/ && return fcol 'pn','gu',$ts;
  /0lu/ && return fcol 'np','lu',$ts;


  # next, consider one sign and one (non-strict) inequality.
  (/(.?)p(.?)g(.?)/) && return fcol 'n',$1.$2.$3,$ts;
  (/(.?)n(.?)l(.?)/) && return fcol 'p',$1.$2.$3,$ts;


  # next, we have cases that imply  inequalities
  /(.?)ll(.?)/ && return fcol 'g',$1.$2,$ts;
  /(.?)gg(.?)/ && return fcol 'l',$1.$2,$ts;

  # one zero, one-nonzero implies the third is non-zero
  (/0(.?)u(.?)/) && return fcol 'u',$1.$2,$ts;

  # some of the remaining cases would require creating disjunctive
  # constraints.
  return 'q';		
}

sub merge_sign($$){
  my ($a,$b)=@_;

  return $sign_short{$encoding{$a} & $encoding{$b}};
}

sub compatible_signs($$){
  my ($a,$b)=@_;

  return (merge_sign($a,$b) ne 'c');
}

sub enforce($@){
  my $settings=shift;
  my @row=@_;

  my ($istr,$sigstr)=split(/:/,$settings);
  
  my @indices=split('',$istr);
  my @signs=split('',$sigstr);
  my @return=();
  for my $i (@indices){
    my $sigma=shift(@signs);
    $sigma=$neg{$sigma}       if ($i == 1);

    my @actions=@{enforce_pair($i,$sigma,$row[2*$i],$row[2*$i+1])};
    push(@return,@actions);
  }
  return @return;
}

sub sign_subset($$){
  my ($sigma,$set)=@_;

  return (($encoding{$sigma} & $encoding{$set}) != 0);
}

sub subtract_sign($$){
  my ($sigma,$set)=@_;
  my $k="$encoding{$set}" & "".complement($encoding{$sigma});
  
  return $sign_short{$k};

}
sub real_enforce_pair(@);
my %ep_cache=();
sub enforce_pair($$$$){
  return cache_call(\&real_enforce_pair,%ep_cache,@_);
}

sub real_enforce_pair(@){
  my ($triple,$sigma,@chi)=@_;

  my(@orig_chi)=@chi;

  my $current=mult($chi[0],$chi[1]);

  sign_subset($sigma,$current) || 
    die("Impossible enforce_pair request $sigma, $chi[0] * $chi[1]");

  return () if ($current eq $sigma);

  if (!sign_subset('0',$sigma)){
    @chi = map {$_ = subtract_sign('0',$_)} @chi;
  }

  my @candidates=();

  for (my $offset=0;$offset < 2; $offset++){

    my $sigma2=$chi[($offset+1)%2];

    for my $sigma3 ( @signs ){
      next if ($sigma3 eq $chi[$offset]);
      next if (!sign_subset($sigma3,$chi[$offset]));

      my $result=mult($sigma2,$sigma3);

      if ($sigma eq $result){
	push(@candidates, "$offset:$sigma3");
      }
    }
  }

  # only if the operation is forced (i.e. unique), do we apply it.
  if (scalar(@candidates) == 1){
    my ($i,$s)=split(":",$candidates[0]);
    $chi[$i]=$s;
  }

  my @returns;

  for (my $i=0; $i<2; $i++){
    my $index=2*$triple+$i;
    if ($chi[$i] ne $orig_chi[$i]){
      push(@returns,"$index:$chi[$i]");
    }
  }

  return [@returns];
}


my $mask='';
my %count=map ( ($_ => 0), ('t','f','S','q','Q'));
my %found=();
my $max_forced=0;
my $token=0;
my $check_rows= 2**($encoding_width*$bases);
my $MAX_TOKENS=255;
print "#undef DOXYGEN_SKIP\n";
print "#ifndef DOXYGEN_SKIP\n";
print "#define PLUECKER_CHECK_ROWS $check_rows\n";
print "pl_token_t pluecker_check[PLUECKER_CHECK_ROWS]={\n";

for (my $count=0; $count < $check_rows; $count++){
  vec($mask,0,$bitwidth) = $count;
  my $bits=substr(unpack("B$bitwidth",$mask),$bitwidth-$encoding_width*$bases);

  $bits=join("",reverse(unpack("a3a3a3a3a3a3",$bits)));

  print STDERR  "." if ($count %1000 ==0 );


  my @row=();
  my @terms=();
  for (my $j=0; $j<$bases; $j++){
    my $b=$sign_short{substr($bits,$encoding_width*$j,$encoding_width)};
    push (@row,$b);
  }
  for (my $j=0; $j<$bases/2; $j++){
    my $t=mult($row[2*$j],$row[2*$j+1]);
    $t=$neg{$t} if ($j==1);
    push(@terms,$t);
  }

  my $settings=pluecker_check(@terms);
  my @actions=();
  my $key="";
  if (!($settings =~ /^[tfq]$/)){
    @actions=enforce($settings,@row);
    if (scalar(@actions)>0) {
      $count{S}++;
      $key=join(',', sort @actions);
      $max_forced=scalar(@actions) if (scalar(@actions)>$max_forced);
    } else {
      $key="t";
      $count{Q}++;
    }
  } else {
    $count{$settings}++;
    if ($settings eq 'f'){
      $key='f';
    } else {
      $key="t";
    }
  }
  if (!defined($found{$key})){
    $found{$key}=$token++;
    ($token < $MAX_TOKENS) || die;
  }
  my $r=join("",@row);
  print "$found{$key},";
  print "\t/* $r\t$key */" if ($debug);
  print "\n";
  
}

print "};\n";


print STDERR "\ndistinct actions = ",scalar(keys(%found)),"\n";
print STDERR "max forced = ",$max_forced,"\n";

for my $s (keys %count){
  print STDERR "$s $count{$s}\n";
}

printf "#define PL_ACTION_ROWS %d\n",$token;

print "pl_action_t pl_actions[PL_ACTION_ROWS]={\n";
my %actions_for=map( ($found{$_}=>$_)  , keys(%found));
print "{ 0 , }, { 0 ,  },\n";
for (my $t=2; $t < $token; $t++){
  my $actions=$actions_for{$t};
  my @actions=split(",",$actions);
  print "{ ";
  printf "%d,",scalar(@actions);
  print "{ ";
  for my $a (@actions){
    my ($offset,$sign)=split(":",$a);
    printf "{ $offset, $def_name{$sign} },";
  }
  print "}},\n";
#  print  "/*$t\t$actions*/\n";
}

print "};\n";
printf "#endif\n";
