#include <stdlib.h>
#include <stdio.h>
#include "int_queue.h"
#include "error_message.h"
#include "nalloc.h"
#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif
#include "memalloc.h"

na_node_p iq_nalloc(na_node_p node, slot_t slot, natural_t qsize){
 int_queue_p new;

 /* adjust for wasted element */

 qsize++;
  node=na_alloc(node,slot,sizeof(int_queue_t)+qsize*sizeof(int));
  new=na_access(node,slot);
  new->size=qsize;
  new->head=0;
  new->tail=0;
  return node;
}


int_queue_p iq_new(natural_t qsize){
  int_queue_p new;

  /* allow room to actually store qsize elements */
  qsize++; 
  new=memalloc(sizeof(int_queue_t)+qsize*sizeof(int));
  new->size=qsize;
  new->head=0;
  new->tail=0;
  return new;
}

void iq_destroy(int_queue_p p){
  free(p);
}

int iq_empty(int_queue_p Q){
  return (Q->head == Q->tail);
}

int iq_full(int_queue_p Q){
  return ( ( (Q->tail +1) % Q->size) == Q->head);
}

int iq_count(int_queue_p Q){
  return (Q->tail+Q->size-Q->head) %Q->size;
}

int iq_deq(int_queue_p Q){
  int r;

  if (iq_empty(Q)){
    err_fatal(HERE,"Queue empty");
  }
  r=Q->elem[Q->head];
  Q->head= (Q->head+1) % Q->size;
  return r;
}

void iq_flush(int_queue_p Q){
  Q->head=0;
  Q->tail=0;
}

int iq_first(int_queue_p Q){
  int r;

  if (iq_empty(Q)){
    err_fatal(HERE,"Queue empty");
  }
  r=Q->elem[Q->head];
  return r;
}

void iq_enq(int_queue_p Q, int i){
  if (iq_full(Q)){
    err_fatal(HERE,"Queue full");
  }

  Q->elem[Q->tail]=i;
  Q->tail = (Q->tail+1) % Q->size;
}

void iq_dump(int_queue_p Q){
  int i;
  for (i=Q->head; (i % Q->size) != Q->tail; i=(i+1) % Q->size ){
    fprintf(stderr,"%d\n",Q->elem[i]);
  }
}
