#ifndef _ERROR_MSG_H
#define _ERROR_MSG_H 1

#include <stdarg.h>
#include <stdio.h>

#ifdef HAVE_CONFIG_H
#include "util_config.h"
#endif

/* 
 * Preprocessor sillyness to define a macro HERE that points to the 
 * current file and line number
 */

#define __here(l) __FILE__":"#l
#define _here(l) __here(l)
#define HERE _here(__LINE__)

#define ERR_TRAP_NULL(p) if (p==NULL){ perror(HERE); abort(); };

/* how many bits are reserved by this file */
#define ERR_MASK_OFF 2

#define ERR_MASK(b) (1<<(b+ERR_MASK_OFF))

#define DB_RETURNS 1
#define DB_RETURNS_ALL 2

#ifdef ERR_NO_DEBUG
#define err_debug(m,a) 
#else 
#define err_debug(m,a) if ((ERR_DEBUG_MASK) & (m)) err_warning a
#endif


void err_fatal(char *where,char *fmt, ...);
void err_warning( char *where, char *fmt, ...);

#ifndef ERR_NO_ASSERT
#define err_assert1(loc,EX) EX ? ((void)0) :err_fatal(loc, "assertion failed: %s",#EX)
#define err_assert(EX) err_assert1(HERE,EX)
#else 
#define err_assert1(loc,EX) ((void)0)
#define err_assert(EX) ((void)0)
#endif


#define ERR_RETURN(x) { \
	if (((ERR_DEBUG_MASK&DB_RETURNS) && (x==0)) || \
	    (ERR_DEBUG_MASK&DB_RETURNS_ALL) ) \
  err_warning(HERE,"returning %x",x); \
  return x; }

#endif

