/**
 * @file   path.c
 * @author David Bremner
 * @date   Wed Apr 17 14:44:32 2002
 * 
 * 
 * @brief generate all inclusion minimal paths of a given length in
 * the pivot complex.
 * 
 * These are the paths that are at least geodesics in the subgraph of
 * the pivot graph induced by their vertices.
 *	   
 * This is the library-version of allpath.c
 *  
*/



#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "oms_error.h"
#include "oms_types.h"
#include "basis_equation.h"
#include "statearray.h"
#include "path.h"

void path_bt(int path[], int step,
	     int limit,
	     int sink,
	     be_block_p pivots, sarray_p found,
	     path_hook found_hook){
  
  int i;

  int parent;

  int prune=0,complete=0;
  int source;

  if (step >= limit) 
    return;

  source=path[step];

  sa_save(found);

  sa_set(found,source,1);

  if (step>0)
    parent=path[step-1];
  else
    parent=-1;



  for (i=0; i<be_cols(pivots); i++){
    array_index_t nbr=be_bindex(pivots,source,i);
    if (!sa_read(found,nbr)){
      if (nbr==sink){
	complete=1;
      }
    } else {
      if (nbr != parent){
	prune=1;
      }
    }

  }

  if (!prune && complete){
    found_hook(path,step+1); 
  }

  if (!prune && !complete){
    for (i=0; i<be_cols(pivots); i++){
      array_index_t nbr=be_bindex(pivots,source,i);
      path[step+1]=nbr;
      if (nbr != parent)
	path_bt(path,step+1,limit,sink,pivots,found,found_hook);
    }
  }

  
  sa_restore(found);
}



    
