/* based on  z_bb4.c   Originally by Ambros Marzetta.

   This program strongly depends on the checkpoint file format defined in
   z_lb.c 


 * $Source$
 * $Revision: 3145 $
 * $Date: 2004-08-25 18:01:45 -0300 (Wed, 25 Aug 2004) $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
#include "zram_config.h"
#endif

#include "z_bb.h"
#include "z_lb.h"		/* z_lb_CheckpointPath */
#include "z_io.h"
#include "z_ps.h"
#include "s_seq.h"
#include "z_checkpoint.h"
#include "z_checkpoint_file.h"
#include "z_arch.h"

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

zconf_p z_conf; /* used to return config from yacc */
extern FILE *zcin;




/* Static variables ----------------------------------------------------- */

static StatArrayT stat;
static boolean check_inited;

static z_JobNum CheckpointParamJobNum;

/* Static procedures */
static void BackupCheckpoint(zconf_p);
static void RewriteCheckpoint(zconf_p zcp,s_seqP workList);
static void z_init_zconf(zconf_p zcp, size_t n_procs, const char *outname);
static zconf_p z_rebuild_zconf(zconf_p zcp, size_t n_procs, const char *outname);
static void CheckpointParamJob(void);
static size_t zconf_size(int numprocs);
static void z_check_broadcast(zconf_p zcp);
static zconf_p z_check_readConfig(const char *name);
static void z_pathcpy(char *buf,const char *name,const char *suffix);
static void WriteGlobals(const z_NodeP, const char*);
/* Procedures ----------------------------------------------------------- */


void z_check_writeConfig(zconf_p zcp){
  FILE *fp;
  int i;

  if ((fp=fopen(zcp->name,"w"))==NULL)
    {
      perror("");
      z_fatal(HERE,"fopen/w %s",zcp->name);
    }
  fprintf(fp,"num_procs: %d\n",zcp->numprocs);
  fprintf(fp,"version: %d\n",zcp->file_version);
  fprintf(fp,"seq: %d\n",zcp->seq);
  fprintf(fp,"sentinel: 0x%x\n",zcp->sentinel);
  fprintf(fp,"arch: %s\n",zcp->arch);




  fprintf(fp,"global { checkpoint: %s }\n",
	  zcp->global.checkpoint);

  for (i=0; i<zcp->numprocs; i++){
    fprintf(fp,"proc %d { checkpoint: %s }\n",
	    i,zcp->proc[i].checkpoint);
  }
  fclose(fp);
}

static zconf_p z_check_readConfig(const char *name){

  char buf[PATH_MAX];
  
  z_pathcpy(buf,name,".conf");

  if((zcin=fopen(buf,"r"))==NULL){
    z_perror(HERE,"fopen/r: %s",name);
  }
  
  if (zcparse()!=0)
    z_fatal(HERE,"error parsing checkpoint config %s\n",name);
  
  z_pathcpy(z_conf->name,name,".conf");

  return z_conf;
  
}

s_seqP ReadAllNodes(zconf_p zcp)
{
  int i, j, s, numprocs;
  FILE *file;
  int version, num, checkpoint_end, cp_seq;
  StatArrayT lstat;
  s_seqP workList;

  workList = s_seq_New(s_seq_default);
  
  for (i=0; i<zcp->numprocs; i++) {
    char *from=zcp->proc[i].checkpoint;
    if ((file = fopen(from, "r"))==NULL){
      z_fatal(HERE,"fopen/r: %s\n",from);
    }
    z_fread(&version, sizeof(version), 1, file);
    if (version != zcp->file_version) {
      z_fatal(HERE,"%s: bad checkpoint file version %d\n", from,version);
    }
    z_fread(&numprocs, sizeof(numprocs), 1, file);
    if (numprocs != zcp->numprocs) {
      z_fatal(HERE, "checkpoint file %s is for %d processors\n",
	      from,numprocs);
    }
    z_fread(&cp_seq, sizeof(cp_seq), 1, file);
    if (zcp->seq != cp_seq) {
      z_fatal(HERE,
	      "%s: checkpoint file has sequence number "
	      "%d instead of %d\n",
	      from, cp_seq, zcp->seq);
    }

    z_fread(&lstat, sizeof(lstat), 1, file);
    for (s=0; s<stat_last; s++) stat[s] += lstat[s];
    z_fread(&num, sizeof(num), 1, file);
    z_printf("[%s]in: Nodes %d\n", from, num);
    for (j=0; j<num; j++)
      s_seq_InsertLast(workList, z_ReadNode(file));
    z_fread(&checkpoint_end, sizeof(checkpoint_end), 1, file);
    if (checkpoint_end != CHECKPOINT_END) {
      z_fatal(HERE,"*%s: bad checkpoint file end %x\n", from,checkpoint_end);
    }
    fclose(file);
  } /* for */
  return workList;
} 

void z_check_WriteHeader(FILE *cp_file,
				  int version, int cp_sequence, int numprocs,
				  StatArrayT *stat, int num)
{
  fwrite(&version, sizeof(version), 1, cp_file);
  fwrite(&numprocs, sizeof(numprocs), 1, cp_file);
  fwrite(&cp_sequence, sizeof(cp_sequence), 1, cp_file);
  fwrite(stat, sizeof(StatArrayT), 1, cp_file);
  fwrite(&num, sizeof(num), 1, cp_file);
} /* z_check_WriteHeader */

static void RewriteCheckpoint(const zconf_p zcp,
			      s_seqP workList)
{
  int i, j, numPerFile, numThisFile;

  int numElements;
  FILE *cp_file;
  static StatArrayT lstat;
  z_NodeP workUnit;
  s_seq_refP ref;

  numElements=s_seq_numElements(workList);
  zcp->seq++;

  numPerFile = numElements / zcp->numprocs;
  for (i=0; i<zcp->numprocs; i++) {
    numThisFile = numPerFile;
    if (i < numElements % zcp->numprocs) numThisFile ++;
    if ( (cp_file = fopen(zcp->proc[i].checkpoint, "w")) == NULL){
      z_perror(HERE,zcp->proc[i].checkpoint);
    }
    
    z_check_WriteHeader(cp_file, zcp->file_version, zcp->seq, zcp->numprocs,
			  &lstat, numThisFile);
    fprintf(stderr, "[%s]out: Nodes %d\n", zcp->proc[i].checkpoint, numThisFile);
    for (j=0; j<numThisFile; j++) {
      ref = s_seq_First(workList);
      workUnit = s_seq_GetObject(workList, ref);
      z_WriteNode(cp_file, workUnit);
      s_seq_Delete(workList, ref);

      z_FreeNode(workUnit);
    }
    fwrite(&(zcp->sentinel), sizeof(zcp->sentinel), 1, cp_file);
    fclose(cp_file);
  } /* for i */
  z_check_writeConfig(zcp);

} /* WriteAll */

static void z_pathcpy(char *buf,const char *name,const char *suffix){
  strncpy(buf,name,PATH_MAX);
  strncat(buf,suffix,PATH_MAX);
}

void z_check_saveConfig(int cp_sequence){
  z_conf->seq=cp_sequence;
  z_check_writeConfig(z_conf);
}
static zconf_p z_rebuild_zconf(zconf_p zcp, size_t n_procs, const char *outname){

  if (n_procs != zcp->numprocs){
    zcp=realloc(zcp,zconf_size(n_procs));
  }
  
  z_init_zconf(zcp,n_procs,outname);
  return zcp;
}

static void z_init_zconf(zconf_p zcp, 
			 size_t n_procs, 
			 const char *outname
	){
  int i;


  zcp->numprocs=n_procs;
  zcp->file_version=CHECKPOINT_VERSION;
  zcp->sentinel=CHECKPOINT_END;
  z_pathcpy(zcp->name,outname,".conf");
  z_pathcpy(zcp->global.checkpoint,outname,".globals");
  
  strncpy(zcp->arch,ZRAM_ARCH_STR,STRSIZ);
    
  for(i=0; i<n_procs; i++){
    char buf[PATH_MAX];
    sprintf(buf,".%03d",i);
    z_pathcpy(zcp->proc[i].checkpoint,outname,buf);
  }
  
}


zconf_p z_check_redist(zconf_p zcp, const char* outname)
{
  s_seqP workList;
  int n_old,n_new;

  assert(check_inited);

  n_old=zcp->numprocs;
  n_new=z_numprocs;

  assert(n_old >0 && n_new >0);
  
  workList=ReadAllNodes(zcp);
  BackupCheckpoint(zcp);

  zcp=z_rebuild_zconf(zcp,n_new,outname);

  RewriteCheckpoint(zcp,workList);
  
  return zcp;
} 


static void PrintStatistics(void)
{
  printf("Total WorkAndSplit %ld %ld (redundant)\n", stat[stat_wsnonred],
	 stat[stat_wsred]);
  printf("Total Requests generated %ld\n", stat[stat_or]);
  printf("Total Requests forwarded %ld\n", stat[stat_rf]);
  fflush(stdout);
} /* PrintStatistics */

/* Rename nproc files name.* to name.*.bak */
/* return -1 if one of them doesn't exist */

void z_check_BackupFile(const char *from){
  char to[PATH_MAX];
  z_pathcpy(to,from, ".bak");
  
  if (rename(from, to)) {
    if (errno != ENOENT) 
      z_perror(HERE,"rename: %s %s",from, to);
  }
}

static void BackupCheckpoint(zconf_p zcp)
{

  int i;

  z_check_BackupFile(zcp->global.checkpoint);
  z_check_BackupFile(zcp->name);
  for (i=0; i<zcp->numprocs; i++) {
    z_check_BackupFile(zcp->proc[i].checkpoint);
  }

} /* z_check_backup */


/* 
   restore global info, return it.
   reorganize the checkpoint data for the current number of
   processors
*/

z_NodeP z_check_restart(const char *confname, const char *outname,
			int timeout){
  z_NodeP globalNode;
  FILE *gf;

  z_conf=z_check_readConfig(confname);

  if ((gf=fopen(z_conf->global.checkpoint,"r"))==NULL)
    z_fatal(HERE,"fopen/r: %s",z_conf->global.checkpoint);

  globalNode=z_ReadNode(gf);
  fclose(gf);  

  z_conf=z_check_redist(z_conf,outname);

  WriteGlobals(globalNode,z_conf->global.checkpoint);
  z_check_broadcast(z_conf);
  if (timeout>0)
    z_StartCheckpointTimer(timeout);

  return (globalNode);
}

void z_check_setup(z_NodeP globals, const char *outname,int timeout){

  z_conf=malloc(zconf_size(z_numprocs));
  z_init_zconf(z_conf,z_numprocs,outname);


  WriteGlobals(globals,z_conf->global.checkpoint);
  z_check_broadcast(z_conf);
  z_check_writeConfig(z_conf);
  
  if (timeout>0)
    z_StartCheckpointTimer(timeout);
}

static void WriteGlobals(const z_NodeP globals, const char *name){
  FILE *gf;

  if ((gf=fopen(name,"w"))==NULL){
    z_perror(HERE,name);
  }

  z_WriteNode(gf,globals);
  fclose(gf);

}
void z_check_Init(void)
{
  /* called on every processor by initproc() */
  if (!check_inited) {
    check_inited = 1;
    z_io_Init();
    CheckpointParamJobNum=z_InstallJob(CheckpointParamJob);

  }
} /* z_check_Init */





/*  Set checkpoint file name on all processors */

static void CheckpointParamJob(void){
  int nprocs;


  nprocs=z_GetMINTBuffer();
  z_conf=malloc(zconf_size(nprocs));

  z_GetDataBuffer(z_conf,zconf_size(nprocs));
  strncpy(z_lb_CheckpointPath,z_conf->proc[z_myproc].checkpoint,PATH_MAX);
}


static size_t zconf_size(int n_procs){
  return sizeof(zconf_t)+(n_procs-1)*sizeof(zconf_proc_t);
}

static void z_check_broadcast(zconf_p zcp){
  size_t len;

  assert(z_frontend);
  assert(check_inited);

  len=zconf_size(zcp->numprocs);

  z_AddMINTBuffer(zcp->numprocs);
  z_AddDataBuffer(zcp,len);
  z_StartParallelJob(CheckpointParamJobNum,NULL);
  strncpy(z_lb_CheckpointPath,zcp->proc[0].checkpoint,PATH_MAX);
  z_WaitWork();
}

