/* z_par.c (parallel base functions)

   Ambros Marzetta
   94-11-10

 * $Source$
 * $Revision: 2863 $
 * $Date: 2002-04-08 04:54:27 -0300 (Mon, 08 Apr 2002) $
 * $Log$
 * Revision 1.4  2002/04/08 07:54:27  bremner
 * fixing some memory bugs.  Also filled in some static initialization values.
 *
 * Revision 1.3  2002/04/03 11:50:54  bremner
 * tracking ugly memory leaks
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.12  1998/03/10  15:02:40  marzetta
 * Added an assertion.
 *
 * Revision 1.11  1996/06/03  19:00:48  marzetta
 * New definition of z_NewNode.
 *
 * Revision 1.10  1996/04/09  13:27:12  marzetta
 * Added SEQUENTIAL, Cenju-3, new install procedures
 *
 * Revision 1.9  1995/09/14  12:27:52  marzetta
 * AddDataBuffer, GetDataBuffer
 *
 * Revision 1.8  1995/08/22  12:39:45  marzetta
 * New message passing, MPI
 *
 * Revision 1.6  1995/07/28  14:05:58  marzetta
 * Some cosmetic changes
 *
 * Revision 1.5  1995/07/28  08:35:28  marzetta
 * struct buffer_header, Music compatibility
 *
 * Revision 1.4  1995/07/27  12:20:08  marzetta
 * Corrected error with headersize
 *
 * Revision 1.3  1995/07/27  11:36:00  marzetta
 * Separation of message passing interface
 *
 * Revision 1.2  1995/07/24  13:58:43  marzetta
 * *** empty log message ***
 *
 */

#include <stdio.h>
#include <stdlib.h>		/* abort */
#include <string.h>
#include "z_par_intern.h"
#ifdef PEonly
#include <musiclib.h>
#endif /* PEonly */
#include "z_md.h"

#define TRACESEND0
#define TRACERECEIVE0
#define TRACERECEIVE10
#define TRACETOKEN0
#define TRACEIDLE0
#define TRACECALLHANDLER0

static boolean z_par_inited=0;

boolean receiving;

/* Buffers ----------------------------------------------------------- */

static long headersize;		/* size of z_Node header without pointers */
extern struct buffer *inbuffer, *outbuffer; /* allocated by z_md_Init */
long inpos, outpos;

/* Node numbering ---------------------------------------------------- */

int z_numprocs;
/* Number of processing elements. They are numbered from 0..z_numprocs-1.
   Processor number 0 always exists.
   There may be an additional processor (front end) with number -1.
   */

int z_myproc;
/* A number from -1..z_numprocs-1. */

boolean z_frontend;
/* Does this process run on the front end (Music) or on processor 0 (Paragon)?
   z_frontend returns true on exactly one processor. If there is a processor
   with number -1, it is the front end. Otherwise processor 0 is front end and
   processing element. (see z_WaitWork()) */

/* Buffers ----------------------------------------------------------- */

#ifndef SEQUENTIAL
extern long z_bufsize;

typedef struct z_FamilyNode {
  Z_NODE_HEADER;
  z_NodeP ch[10];
} z_FamilyNodeT, *z_FamilyNodeP;

void z_AddDataBuffer(const void *source, long length)
{
  z_buf_maybe_realloc(outpos * sizeof(unitT) + length);
  assert(outpos * sizeof(unitT) + length <= z_bufsize);
  (void) memcpy(outbuffer->b+outpos, source, length);
  outpos += UNITS(length);
} /* z_AddToBuffer */

void z_AddTagBuffer(MINT tag)
{
  outbuffer->header.msg_tag = tag;
}

void z_AddMINTBuffer(MINT x)
{
  z_buf_maybe_realloc(outpos * sizeof(unitT) + sizeof(MINT));
  assert(outpos * sizeof(unitT) + sizeof(MINT) <= z_bufsize);
  (void) memcpy(outbuffer->b+outpos, &x, sizeof(MINT));
  /*printf("Add MINT %ld pos = %ld\n", (long)x, (long)outpos);*/
  outpos += UNITS(sizeof(MINT));
}

void z_AddNodeBuffer(z_NodeP source)
{
  /* Copy source to outbuffer. source->len is replaced: the length is
     measured in other units.
     */
  z_FamilyNodeP header, src;
  char *body;  /* body = node without header and children pointers */
  long bodylen;			/* length of body */
  long newlen;			/* length of body in unitT units */
  unsigned int i;

  if (source) {
    z_AddMINTBuffer(1);
    src = (z_FamilyNodeP) source;
    /*printf("[%2d]AddNode tag=%ld len=%ld children=%ld pos=%ld\n", z_myproc, (long)src->tag, (long)src->len, (long)src->children, outpos);*/
    body = (char *) & src->ch[src->children];
    bodylen = ((long) src) + source->len - ((long) body);
    assert(bodylen >= 0);
    newlen = UNITS(bodylen);

    z_buf_maybe_realloc(outpos * sizeof(unitT) + headersize + bodylen);
    assert(outpos * sizeof(unitT) + headersize + bodylen <= z_bufsize);
    { /* copy the header and replace the length */
      (void) memcpy(outbuffer->b+outpos, source, headersize);
      header = (z_FamilyNodeP) (outbuffer->b+outpos);
      header->len = newlen;
      outpos += UNITS(headersize);
    }
    { /* copy the body */
      /*printf("body: %ld, %ld\n", outpos, bodylen);*/
      (void) memcpy(outbuffer->b+outpos, body, bodylen);
      outpos += newlen;
    }
    /*printf("%ld}\n", outpos);*/
    { /* copy the children */
      for (i=0; i<src->children; i++) z_AddNodeBuffer(src->ch[i]);
    }
  } else z_AddMINTBuffer(0);
} /* z_AddNodeBuffer */

MINT z_GetTagBuffer(void)
{
  assert(! receiving);
  return inbuffer->header.msg_tag;
}

void z_GetDataBuffer(void *dest, long length)
{
  memcpy(dest, inbuffer->b+inpos, length);
  inpos += UNITS(length);
} /* z_GetDataBuffer */

MINT z_GetMINTBuffer(void)
{
  MINT x;

  assert(! receiving);
  x = * (MINT *) (inbuffer->b+inpos);
  /*printf("[%2d]Get MINT: pos %5ld: MINT %ld\n", z_myproc, inpos, (long)x);*/
  inpos += UNITS(sizeof(MINT));
  return x;
} /* z_GetMINTBuffer */

z_NodeP z_GetNodeBuffer(void)
{
  z_FamilyNodeP header, np;
  z_FamilyNodeT header_struct;
  long locallen;
  boolean tag;
  long bodylen;			/* length of body */
  
  assert(! receiving);
  tag = z_GetMINTBuffer();	/* NULL or not NULL? */
  if (tag) {
    /* Added memcpy to avoid alignment problems, db */
    
    header =&header_struct;
    (void)memcpy(header,inbuffer->b+inpos,sizeof(z_FamilyNodeT));
    /*    header = (z_FamilyNodeP) (inbuffer->b+inpos); */
    bodylen = header->len * sizeof(unitT);
    np = NULL;
    /*    locallen = ((long) &np->ch[header->children]) + bodylen - ((long) np); */
        locallen =(long)(np->ch+header->children) +bodylen; 
	/*    printf("[%2d]GetNode tag=%ld, len=%ld, children = %ld pos = %ld\n", z_myproc, (long)header->tag, (long)locallen, (long)header->children, inpos); */
    np = (z_FamilyNodeP) z_NewNode(header->tag, locallen);
    assert(np);
    { /* copy the header and replace the length */
      (void) memcpy(np, inbuffer->b+inpos, headersize);
      np->len = locallen;
      inpos += UNITS(headersize);
    }
    { /* copy the body */
      /*printf("body: %ld, %ld\n", inpos, bodylen);*/
      (void) memcpy(&np->ch[np->children], inbuffer->b+inpos, bodylen);
      inpos += UNITS(bodylen);
    }
    /*printf("%ld}\n", inpos);*/
    { /* copy the children */
      unsigned int i;
      for (i=0; i<np->children; i++) np->ch[i] = z_GetNodeBuffer();
    }
    {
      np->parent = NULL;
      np->child = NULL;
      np->sibling = NULL;
    }
    return (z_NodeP) np;
  } else
    return NULL;
} /* z_GetNodeBuffer */
     
/* Starting jobs ---------------------------------------------------- */

#define MAXJOB 50
static z_VoidProc gjob[MAXJOB];
static const char *jobName[MAXJOB];
static z_JobNum numJob=0;

z_JobNum z__InstallJob(const z_VoidProc job, const char * name)
{
  assert(z_par_inited);
  assert(numJob < MAXJOB);
  gjob[numJob] = job;
  jobName[numJob] = name;
  return (numJob ++);
} /* z_InstallJob */

#ifndef PEonly

static boolean job_running;
static z_ResultProc gResultProc;

void z_StartParallelJob(z_JobNum jobnum, z_ResultProc resultProc)
{
  /* Broadcast outbuffer and initiate the parallel computation. */
  assert(z_par_inited);
  assert(z_frontend);
  assert(! job_running);
  assert(jobnum < numJob);
  job_running = 1;
  z_AddTagBuffer(jobnum * 2 + (resultProc ? 1 : 0));
  gResultProc = resultProc;
  z_md_StartParallel();
  /* note that inbuffer and outbuffer are maintained at the same size in 
     z_buf.c */
  if (z_myproc != -1)
    (void) memcpy(inbuffer, outbuffer,
		  outpos * sizeof(unitT) + sizeof(struct buffer_header));
  outpos = 0;
  inpos = 0;
} /* z_StartParallelJob */

void z_WaitWork(void)
{
  /* Called on front end.
     Work if necessary (i.e. if z_myproc == 0:
                        all machines except Music frontend)
     Wait for PE 0 if gResultProc == 0
                       (on most machines the frontend itself is PE 0
		        only with Music the frontend is PE -1)
     Wait for all PEs if gResultProc != 0
                       (all machines except sequential workstation)
     */
  assert(z_frontend);
  assert(job_running);
  if (z_myproc == 0) {			/* work as PE 0 */
    z_pe_switch();
  } /* z_myproc == 0 */
  z_md_Wait(gResultProc);
  job_running = 0;
} /* z_WaitWork */

#else /* ! PEonly */
#if 0
void z_StartParallelJob(z_JobNum jobnum, z_ResultProc resultProc)
{ assert(0); /* never called, but linker needs it */}

void z_WaitWork(void)
{ assert(0); /* never called, but linker needs it */}
#endif
#endif /* ! PEonly */

/* PE loop -------------------------------------------------------------- */

void z_pe_switch(void)
{
  /* Called by main (Paragon/A7), z_WaitWork and MusicStartParallel */
  /* Pre: There is a message in inbuffer */
  /* Post: All processors have sent their result	(if collect)
	   The frontend has its result in inbuffer	(if not collect).
	   outpos = 0.
	   */
  int parproctag;
  boolean collect;
  
  assert(z_myproc != -1);
  parproctag = z_GetTagBuffer();
  collect = parproctag % 2;
  parproctag /= 2;
  /*printf("[%2d]Received parproctag %ld\n", z_myproc, (long) parproctag);*/
  assert(parproctag >= 0 && parproctag < (int)numJob);
  /*printf("[%2d]Starting job %s\n", z_myproc, jobName[parproctag]);*/
  inpos = 0;
  gjob[parproctag]();
  assert(! receiving);
#ifndef PEonly
  if (z_frontend) {
    (void) memcpy(inbuffer, outbuffer,
		  outpos * sizeof(unitT) + sizeof(struct buffer_header));
    inpos = 0;			/* inbuffer is full */
    if (gResultProc)		/* collect result from PE 0 */
      gResultProc(0);
  } else
#endif /* PEonly */
    z_md_SendToFE(collect);
  outpos = 0;
} /* z_pe_switch */

static z_JobNum terminate_jobnum;
static boolean terminate;
static void TerminateJob(void)
{
  terminate = 1;
} /* TerminateJob */

void z_pe_loop(void)
{
  /* Called by main() and MusicSlave() */
  assert(! z_frontend);
  while (!terminate) {
    z_md_RecBuffer();
    z_pe_switch();
  } /* while (! terminate) */
} /* z_pe_loop */
#endif /* SEQUENTIAL */

/* Assert ---------------------------------------------------------- */

void z_assert(const int proc, const char *ex,
	      const char *file, const int line)
{
  /* A7: Should consume all messages to avoid deadlock and send error
   message to PE0.
   */
#ifndef A7
  fprintf(stderr, "*%2d*Assertion failed: %s, file %s, line %d\n",
	  proc, ex, file, line);
#else /* A7 */
  printf("*%2d*Assertion failed: %s, file %s, line %d\n",
	  proc, ex, file, line);
#endif /* A7 */
  z_md_abort();
} /* z_assert */

/* Initialization --------------------------------------------------- */

void z_par_Init(int myproc, int numprocs, boolean frontend)
/* Initialize this module. Called by z_md_Init() and MusicInitNode(). */
{
  assert(!z_par_inited);
  if (!z_par_inited) {
    z_par_inited = 1;

    z_myproc = myproc;
    z_numprocs = numprocs;
    z_frontend = frontend;

    assert(1 <= numprocs && numprocs <= MAX_PE);
    assert(-1 <= myproc && myproc < numprocs);
#ifndef SEQUENTIAL
    assert(inbuffer);
    assert(outbuffer);
    terminate_jobnum = z_InstallJob(TerminateJob);
#endif /* SEQUENTIAL */

    receiving = 0;
    headersize = (long)&((z_NodeP)NULL)->parent;
    if ((long)&((z_NodeP)NULL)->sibling < headersize)
      headersize = (long)&((z_NodeP)NULL)->sibling;
  }
} /* z_par_Init */

void z_par_Fin(void)
{
  /* Called by main() */
#ifndef PEonly
#ifndef SEQUENTIAL
  if (z_frontend)		/* Terminate other processes. */ 
    z_StartParallelJob(terminate_jobnum, NULL);
#endif /* SEQUENTIAL */
#endif /* PEonly */
  z_md_Fin();
} /* z_par_Fin */
